"""Converters to and from multi-valued decision diagrams."""

from functools import reduce
from typing import Optional

import aiger_bv as BV
import funcy as fn
import mdd
from aiger_bv.bundle import Bundle

from aiger_discrete import Encoding, FiniteFunc


def cone(circ: BV.AIGBV, output: str) -> BV.AIGBV:
    """Return cone of influence aigbv output."""
    for out in circ.outputs - {output}:
        size = circ.omap[out].size
        circ >>= BV.sink(size, [out])
    assert len(circ.outputs) == 1
    assert fn.first(circ.outputs) == output
    return circ


def to_var(bdl: Bundle, encoding: Optional[Encoding]) -> mdd.Variable:
    if encoding is None:
        encoding = Encoding()

    return mdd.Variable(
        encode=encoding.encode,
        decode=encoding.decode,
        valid=BV.uatom(bdl.size, bdl.name)[0] | 1,  # const 1.
    )


def onehot_output(expr):
    """Creates circuit that depends only on 1-hot active bit."""
    bits = BV.uatom(expr.size, expr.output)

    def ite(test, idx):
        return BV.ite(expr[idx], bits[idx], test)

    # Create chained if then else testing 1-hot bit.
    return reduce(ite, range(1, expr.size), bits[0])


def to_mdd(func: FiniteFunc, manager=None) -> mdd.DecisionDiagram:
    # TODO: currently assuming 1-hot.
    assert len(func.outputs) == 1
    output = fn.first(func.outputs)
    expr = BV.UnsignedBVExpr(cone(func.circ, output))
    output = to_var(func.circ.omap[output], func.output_encodings.get(output))

    imap = func.circ.imap
    input_encodings = func.input_encodings
    inputs = {
        k: to_var(imap[k], input_encodings.get(k)) for k in func.circ.inputs
    }

    interface = mdd.Interface(
        inputs=inputs,
        output=output,
        valid=BV.UnsignedBVExpr(cone(func.circ, func.valid_id)),
    )

    return interface.lift(onehot_output(expr))



"""Functions/Classes for modeling Finite Functions as And Inverter Graphs."""

from __future__ import annotations

import re
from functools import reduce
from typing import Any, Callable, Optional, Union
from uuid import uuid1

import aiger_bv as BV
import aiger_ptltl as LTL
import attr
import funcy as fn
from pyrsistent import pmap
from pyrsistent.typing import PMap


TIMED_NAME = re.compile(r"(.*)##time_\d+$")


def fresh():
    return str(uuid1())


@attr.s(auto_attribs=True, frozen=True)
class Encoding:
    """Encodes/decodes finite domain to and from bit sequences (ints)."""
    encode: Callable[[Any], int] = fn.identity
    decode: Callable[[int], Any] = fn.identity


DefaultEncoding = Encoding()
Encodings = PMap[str, Encoding]


@attr.s(auto_attribs=True, frozen=True)
class FiniteFunc:
    circ: BV.AIGBV
    input_encodings: Encodings = attr.ib(converter=pmap)
    output_encodings: Encodings = attr.ib(converter=pmap)
    valid_id: str = "##valid"

    def __attrs_post_init__(self):
        if self.valid_id not in self.circ.outputs:
            raise ValueError(f"Missing validation output: {self.valid_id}.")
        elif self.circ.omap[self.valid_id].size != 1:
            raise ValueError("Validation output must be size 1.")

    def rename_valid(self, name: Optional[str] = None) -> FiniteFunc:
        if name is None:
            name = fresh()
        if name in self.outputs:
            raise ValueError(f"{name} conflicts with current outputs.")
        circ = self.circ['o', {self.valid_id: name}]
        return attr.evolve(self, circ=circ, valid_id=name)

    def assume(self, pred: BV.AIGBV) -> FiniteFunc:
        assert len(pred.outputs) == 1
        func = from_aigbv(pred, valid_id=fn.first(pred.outputs))
        return self | func

    @property
    def aigbv(self):
        return self.circ

    @property
    def aig(self):
        return self.circ.aig

    @property
    def _vexpr(self):
        return BV.uatom(1, self.valid_id)

    @property
    def imap(self):
        return self.circ.imap.omit([self.valid_id])

    @property
    def omap(self):
        return self.circ.omap.omit([self.valid_id])

    @property
    def inputs(self): return self.circ.inputs

    @property
    def outputs(self): return self.circ.outputs - {self.valid_id}

    @property
    def latches(self): return self.circ.latches

    @property
    def latch2init(self): return self.circ.latch2init

    simulator = BV.AIGBV.simulator
    simulate = BV.AIGBV.simulate

    def __call__(self, inputs, latches=None):
        inputs = encode_inputs(inputs, self.circ.imap, self.input_encodings)
        omap, lmap = self.circ(dict(inputs), latches=latches)

        valid, *_ = omap[self.valid_id]
        if not valid:
            raise ValueError(f"Invalid inputs: {inputs}")
        del omap[self.valid_id]

        omap = dict(decode_outputs(omap, self.output_encodings))
        return omap, lmap

    def __or__(self, other: Circ) -> FiniteFunc:
        other: FiniteFunc = canon(other)
        circ = (self.circ | other.circ) >> both_valid(self, other)

        # TODO: project to important inputs.
        return from_aigbv(
            circ=circ,
            input_encodings=self.input_encodings + other.input_encodings,
            output_encodings=self.output_encodings + other.output_encodings,
            valid_id=self.valid_id
        )

    def __rshift__(self, other: Circ) -> FiniteFunc:
        other: FiniteFunc = canon(other)
        circ = (self.circ >> other.circ) >> both_valid(self, other)
        return from_aigbv(
            circ=circ,
            input_encodings=other.input_encodings + self.input_encodings,
            output_encodings=self.output_encodings + other.output_encodings,
            valid_id=self.valid_id
        )

    def __lshift__(self, other: Circ) -> FiniteFunc:
        return canon(other) >> self

    def __getitem__(self, others):
        kind, relabels = others
        if kind == 'o' and self.valid_id in relabels:
            raise ValueError("Use rename_valid to change valid_id.")

        kwargs = {}
        if kind == 'o':
            key, encodings = 'output_encodings', self.output_encodings
        elif kind == 'i':
            key, encodings = 'input_encodings', self.input_encodings

        evolver = encodings.evolver()
        for old, new in relabels.items():
            if old not in encodings:
                continue
            evolver[new] = encodings[old]
            del evolver[old]
        kwargs[key] = evolver.persistent()

        return attr.evolve(self, circ=self.circ[others], **kwargs)

    def _encode_wiring(self, wiring):
        if 'init' not in wiring:
            return wiring

        wiring = dict(wiring)  # copy to avoid side-effect.

        name, init = wiring['input'], wiring['init']

        if wiring.get('input_encoding', True):
            encodings = self.input_encodings
        else:
            encodings = self.output_encodings

        if name in encodings:
            init = wiring['encoding'].encode(init)

        assert isinstance(init, int)

        size = self.circ.imap[name].size
        wiring['init'] = BV.encode_int(size, init, signed=False)
        return wiring

    def loopback(self, *wirings):
        wirings = map(self._encode_wiring, wirings)

        return from_aigbv(
            circ=self.circ.loopback(*wirings),
            input_encodings=self.input_encodings,
            output_encodings=self.output_encodings,
            valid_id=self.valid_id,
        )

    def unroll(self, horizon, *, init=True, omit_latches=True,
               only_last_outputs=False):
        hist_valid = LTL.atom(self.valid_id) \
                        .historically() \
                        .with_output(self.valid_id)
        monitor = BV.aig2aigbv(hist_valid.aig)

        circ = (self.circ >> monitor).unroll(
            horizon,
            init=init,
            omit_latches=omit_latches,
            only_last_outputs=only_last_outputs
        )

        if not only_last_outputs:
            times = range(1, horizon)
            circ >>= BV.sink(1, (f'{self.valid_id}##time_{t}' for t in times))
        valid_id = f'{self.valid_id}##time_{horizon}'
        assert valid_id in circ.outputs

        input_encodings = timed_encodings(self.input_encodings, circ.inputs)
        output_encodings = timed_encodings(self.output_encodings, circ.outputs)
        return from_aigbv(
            circ=circ,
            input_encodings=input_encodings,
            output_encodings=output_encodings,
            valid_id=valid_id,
        )


def encode_inputs(inputs, imap, encodings):
    for key, val in inputs.items():
        size = imap[key].size
        if key in encodings:  # Convert to tuple of bools.
            val = encodings[key].encode(val)

        assert isinstance(val, int)
        val = BV.encode_int(size, val, signed=False)

        assert len(val) == size
        yield key, val


def decode_outputs(outputs, encodings):
    for key, val in outputs.items():
        val = BV.decode_int(val, signed=False)
        if key in encodings:
            val = encodings[key].decode(val)
        yield key, val


def timed_encodings(old_encodings, timed_names):
    encodings = {}
    for timed_name in timed_names:
        name = TIMED_NAME.match(timed_name).group()[0]
        if name not in old_encodings:
            continue
        encodings[timed_name] = old_encodings[name]

    return encodings


Circ = Union[FiniteFunc, BV.AIGBV]


def both_valid(left: FiniteFunc, right: FiniteFunc) -> BV.AIGBV:
    return (left._vexpr & right._vexpr).with_output(left.valid_id).aigbv


def canon(circ: Circ) -> FiniteFunc:
    if not isinstance(circ, FiniteFunc):
        circ = from_aigbv(circ)
    return circ.rename_valid()


def omit(mapping, keys):
    return reduce(lambda m, k: m.discard(k), keys, pmap(mapping))


def project(mapping, keys):
    return omit(mapping, set(mapping.keys()) - keys)


def from_aigbv(circ: BV.AIGBV,
               input_encodings: Encodings = None,
               output_encodings: Encodings = None,
               valid_id="##valid") -> FiniteFunc:
    """Lift an bit-vector into a function over finite sets.

    Note: if `valid_id` is not present as an output of `circ`, then it
      will be added, and will always output True.

    Args:
     - input_encodings: Maps an input to an encoder. Default is identity.
     - output_encodings: Maps an output to an encoder. Default is identity.
     - valid_id: Denotes which output monitors if inputs are "valid".
    """
    if input_encodings is None:
        input_encodings = {}
    if output_encodings is None:
        output_encodings = {}
    if valid_id not in circ.outputs:
        circ |= BV.uatom(1, 1).with_output(valid_id).aigbv

    input_encodings = project(input_encodings, circ.inputs)
    output_encodings = project(output_encodings, circ.outputs - {valid_id})

    return FiniteFunc(
        circ=circ,
        input_encodings=input_encodings,
        output_encodings=output_encodings,
        valid_id=valid_id,
    )


__all__ = ['Encoding', 'FiniteFunc', 'from_aigbv']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_discrete']

package_data = \
{'': ['*']}

install_requires = \
['funcy>=1.14,<2.0',
 'py-aiger-bv>=4.3.0,<5.0.0',
 'py-aiger-ptltl>=3.0.1,<4.0.0',
 'pyrsistent>=0.17.3,<0.18.0']

extras_require = \
{'mdd': ['mdd>=0.3.1,<0.4.0']}

setup_kwargs = {
    'name': 'py-aiger-discrete',
    'version': '0.1.2',
    'description': 'Library for modeling functions over discrete sets using aiger circuits.',
    'long_description': "# py-aiger-discrete\nLibrary for modeling functions over discrete sets using aiger circuits.\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-discrete/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-discrete)\n[![docs badge](https://img.shields.io/badge/docs-docs-black)](https://mjvc.me/py-aiger-discrete)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-discrete/branch/main/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-discrete)\n[![PyPI version](https://badge.fury.io/py/py-aiger_discrete.svg)](https://badge.fury.io/py/py-aiger-discrete)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n# Installation\n\nIf you just need to use `py-aiger-discrete`, you can just run:\n\n`$ pip install py-aiger-discrete`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n   \n# About\n\nThis library helps with modeling functions over finite sets using circuits. \n\n    f : A → B₁ × B₂ × … × Bₘ\n\nwhere `A ⊆ A₁ × A₂ × … × Aₙ`.\n\nThis is done by providing:\n\n1. A encoder/decoder pair for each set to and from (unsigned)\n   integers. \n2. A circuit that uses the bit-vector representation of these integers\n   to compute `f`.\n3. A circuit that monitors if the input bit-vector sequence is\n   `valid`, i.e., a member of `A`.\n\nFunctionally, the `py-aiger-discrete` library centers around the\n`aiger_discrete.FiniteFunc` class which has 4 attributes an `aiger_bv.AIGBV` object.\n\n1. A string `valid_id` indicating\n1. A circuit, `circ`, over named bit-vectors in the form of an\n   `aiger_bv.AIGBV` object. One of the outputs must be named\n   `valid_id`.\n1. A mapping from **inputs** to `aiger_discrete.Encoding` objects which\n   encode/decode objects to integers. The standard bit-encoding of\n   unsigned integers is feed into `circ`.\n1. A mapping from **outputs** to `aiger_discrete.Encoding` objects which\n   encode/decode objects to integers. These encodings are used to decode\n   the resulting bit-vectors of `circ`.\n\n# Usage\n\nBelow we provide a basic usage example. This example assumes basic\nknowledge of the `py-aiger` ecosystem and particularly `py-aiger-bv`.\n\n```python\nimport aiger_bv as BV\n\nfrom aiger_discrete import Encoding, from_aigbv\n\n# Will assume inputs are in 'A', 'B', 'C', 'D', or 'E'.\nascii_encoder = Encoding(\n    decode=lambda x: chr(x + ord('A')),  # Make 'A' map to 0.\n    encode=lambda x: ord(x) - ord('A'),\n)\n\n# Create function which maps: A -> B, B -> C, C -> D, D -> E.\n\nx = BV.uatom(3, 'x')  # need 3 bits to capture 5 input types.\nupdate_expr = (x + 1) & 0b111\ncirc = update_expr.with_output('y').aigbv\n\n# Need to assert that the inputs are less than 4.\ncirc |= (x <= 4).with_output('##valid').aigbv\n\nfunc = from_aigbv(\n    circ,\n    input_encodings={'x': ascii_encoder},\n    output_encodings={'y': ascii_encoder},\n    valid_id='##valid',\n)\n\nassert func('A') == 'B'\nassert func('B') == 'C'\nassert func('C') == 'D'\nassert func('D') == 'E'\nassert func('E') == 'A'\n```\n\nNote that `py-aiger-discrete` implements most of the circuit API as `aiger_bv.AIGBV`. \n\nFor example, sequential composition:\n\n```python\nfunc12 = func1 >> func2\n```\n\nor parallel composition:\n\n```python\nfunc12 = func1 | func2\n```\n\nor unrolling:\n\n```python\nfunc_unrolled = func1.unroll(5)\n```\n\nor feedback: \n\n\n```python\nfunc_cycle = func1.loopback({\n    'input': 'x',\n    'output': 'y',\n    'keep_output': True,\n    `input_encoder`: True,\n    `init`: 'A',\n})\n```\n\nNote that feedback now supports additional flag per wiring description\ncalled `input_encoder` which determines if the input or output\nencoding is used for initial latch value resp. The default is the\ninput encoding.\n\n\nor renaming:\n```python\nfunc_renamed = func1['i', {'x': 'z'}]\nassert func1.inputs == {'z'}\n```\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger-discrete',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
